package pathgame.algorithm;

import pathgame.tilemap.TileMap;

import java.util.ArrayList;
import java.util.Collections;

public class TravellingSalesAlg {
    public static int calcSalesPathLen(TileMap map) {
        DijkstraMagic dijkstra = new DijkstraMagic(map);

        ArrayList<ArrayList<SaleRoute>> salesPitch = dijkstra.getSalesPitch();

        //make minimum spanning tree
        ArrayList<TreeEdge> MSTree = makeMSTree(salesPitch);

        //find pairs-shortest path for verteces with odd degree and add those edges to MSTree
        ArrayList<TreeEdge> oddDegEdges = makeOddDegEdges(MSTree, salesPitch);
        MSTree.addAll(oddDegEdges);


        return 10;
    }

    private static ArrayList<TreeEdge> makeOddDegEdges(ArrayList<TreeEdge> msTree, ArrayList<ArrayList<SaleRoute>> salesPitch) {
        int numOfEdges[] = new int[salesPitch.size()];

        for (int i = 0; i < msTree.size(); i++) {
            numOfEdges[msTree.get(i).getSrc()]++;
            numOfEdges[msTree.get(i).getDest()]++;
        }

        OddDegreeList oddDegs = new OddDegreeList();

        for (int i = 0; i < numOfEdges.length; i++) {
            //System.out.println(numOfEdges[i]);
            if (numOfEdges[i] % 2 == 1) {
                oddDegs.add(i);
            }
        }

        Permutation permut = new Permutation(oddDegs.size());

        calcPairShortest(oddDegs, salesPitch, permut, 0, 0);
        permut.makePermutMinimal();

        //System.out.println(permut.tempCounter1);
        //System.out.println(permut.tempCounter2);
        //System.out.println("min cost: " + permut.getMinCost());
        //permut.printPermut();

        ArrayList<TreeEdge> oddEdges = new ArrayList<>();
        oddDegs.resetUsed();
        for(int i = 0; i < permut.size(); i++) {
            int offSet = permut.getValAtPos(i);

            addOddEdge(oddEdges, oddDegs, salesPitch, offSet);
        }

        addOddEdge(oddEdges, oddDegs, salesPitch, 0);

        return oddEdges;
    }

    private static void addOddEdge(ArrayList<TreeEdge> oddEdges, OddDegreeList oddDegs, ArrayList<ArrayList<SaleRoute>> salesPitch, int offSet) {
        int orig = oddDegs.getUnused(0);
        oddDegs.makeUsed(0);
        int dest = oddDegs.getUnused(offSet);
        oddDegs.makeUsed(offSet);

        oddEdges.add(new TreeEdge(orig, dest, salesPitch.get(orig).get(dest - orig - 1).getTotalCost()));
    }

    private static void calcPairShortest(OddDegreeList oddDegs, ArrayList<ArrayList<SaleRoute>> salesPitch, Permutation permut, int permutPos, int costSoFar) {
        while(true) {
            int offSet;

            if(permutPos == permut.size()) {
                offSet = 0;
            }
            else {
                offSet = permut.getValAtPos(permutPos);
            }

            int orig = oddDegs.getUnused(0);
            int dest = oddDegs.getUnused(1 + offSet);

            int edgeWeight = salesPitch.get(orig).get(dest - orig - 1).getTotalCost();
            int newCost = costSoFar + edgeWeight;
            //permut.tempCounter1++;

            /*System.out.println();
            System.out.println("permut pos: " + permutPos);
            System.out.println(orig + " : " + dest);
            System.out.println("newCost: " + newCost);
            permut.printPermut();*/

            if(newCost < permut.getMinCost()) {
                if(permutPos == permut.size()) {
                    permut.setMinCost(newCost);
                    /*System.out.println();
                    System.out.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                    System.out.println("New Min: " + newCost);
                    System.out.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");*/
                }
                else {
                    int used1 = oddDegs.makeUsed(0);
                    int used2 = oddDegs.makeUsed(offSet);
                    calcPairShortest(oddDegs, salesPitch, permut, permutPos+1, costSoFar+edgeWeight);
                    oddDegs.makeUnused(used1);
                    oddDegs.makeUnused(used2);
                    //TODO: do the right stuff
                }
            }

            if(permutPos == permut.size()) {
                //permut.tempCounter2++;
                break;
            }
            if(permut.isPosAtMax(permutPos)) {
                break;
            }

            permut.increaseAtPos(permutPos);



            /*permut.increaseAtPos(3);
            permut.printPermut();
            permut.increaseAtPos(2);
            permut.printPermut();
            permut.increaseAtPos(2);
            permut.printPermut();*/

            //break;
        }
    }

    public static ArrayList<TreeEdge> makeMSTree(ArrayList<ArrayList<SaleRoute>> salesPitch) {
        ArrayList<TreeEdge> allEdges = new ArrayList<>();
        ArrayList<TreeEdge> msTree = new ArrayList<>();

        int vertNum = salesPitch.size();

        for (int orig = 0; orig < salesPitch.size(); orig++) {
            for (int dest = 0; dest < salesPitch.get(orig).size(); dest++) {
                allEdges.add(new TreeEdge(orig, dest + 1 + orig, salesPitch.get(orig).get(dest).getTotalCost()));
                //System.out.println(edge.getSrc() + " " + edge.getDest() + " " + edge.getWeight());
                //System.out.println(salesPitch.get(orig).get(dest).getTotalCost());
            }
        }

        //System.out.println(vertNum);

        Collections.sort(allEdges);

        while (msTree.size() < vertNum - 1) {

            if (notCycle(msTree, allEdges.get(0))) {
                msTree.add(allEdges.get(0));
                //System.out.println(allEdges.get(0).getSrc() + " " + allEdges.get(0).getDest() + " " + allEdges.get(0).getWeight());
            }
            allEdges.remove(0);
        }

        return msTree;
    }

    private static void qEdges(ArrayList<TreeEdge> tree, ArrayList<TreeEdge> edgeQ, int vertex) {
        for(int i = 0; i < tree.size(); i++) {
            TreeEdge edge = tree.get(i);
            if(edge.getSrc() == vertex || edge.getDest() == vertex) {
                if(!edge.isChecked()) {
                    edgeQ.add(edge);
                    //System.out.println("src: " + edge.getSrc() + "; dest: " + edge.getDest());
                    edge.setChecked(true);
                }
            }
        }
    }

    private static boolean notCycle(ArrayList<TreeEdge> tree, TreeEdge additon) {
        ArrayList<TreeEdge> edgeQ = new ArrayList<>();
        int dest = additon.getDest();
        resetEdges(tree);

        qEdges(tree, edgeQ, additon.getSrc());

        while(edgeQ.size() > 0) {
            TreeEdge edge = edgeQ.get(0);
            edgeQ.remove(0);
            if(edge.getSrc() == dest || edge.getDest() == dest) {
                return false;
            }
            else {
                qEdges(tree, edgeQ, edge.getSrc());
                qEdges(tree, edgeQ, edge.getDest());
            }
        }

        return true;
    }

    private static void resetEdges(ArrayList<TreeEdge> tree) {
        for(int i = 0; i < tree.size(); i++) {
            tree.get(i).setChecked(false);
        }
    }
}